#version 400 compatibility

/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 0 */
layout(location = 0) out vec3 sceneColor;


#include "/lib/head.glsl"

in vec2 uv;

flat in mat4x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex3;
uniform sampler2D colortex5;
uniform sampler2D colortex7;
uniform sampler2D colortex8;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform sampler2D noisetex;

uniform int frameCounter;
uniform int worldTime;

uniform float eyeAltitude;
uniform float far, near;
uniform float frameTimeCounter;
uniform float cloudLightFlip;

uniform vec2 taaOffset;

uniform vec3 cameraPosition;
uniform vec3 cloudLightDir, cloudLightDirView;

uniform vec4 daytime;

uniform mat4 gbufferProjectionInverse, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferModelView;

#include "/lib/frag/bluenoise.glsl"

#include "/lib/util/bicubic.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/atmos/project.glsl"

vec3 getFog(vec3 color, float dist, vec3 skyColor, vec3 sceneDir) {

	    dist 	= dist / far;
		dist 	= max((dist - fogStart), 0.0);
        dist   /= 1.0 - fogStart;
        
	float alpha = 1.0-exp(-dist * fogFalloff / euler);
        alpha  *= cube(1.0 - max0(sceneDir.y));

	color 	= mix(color, skyColor, saturate(alpha));

	return color;
}


const float edgeWeightStep = 1.0 - (1.0 / cloudVolumeRounding);

float cloudTextureCubic(sampler2D tex, vec2 pos) {
    ivec2 texSize = textureSize(tex, 0) * cloudVolumeRounding;
    vec2 texelSize = rcp(vec2(texSize));
    
    float p0q0 = texture(tex, pos).a;
    float p1q0 = texture(tex, pos + vec2(texelSize.x, 0)).a;

    float p0q1 = texture(tex, pos + vec2(0, texelSize.y)).a;
    float p1q1 = texture(tex, pos + vec2(texelSize.x , texelSize.y)).a;

    float a = cubeSmooth(fract(pos.x * texSize.x));

    float pInterp_q0 = mix(p0q0, p1q0, a);
    float pInterp_q1 = mix(p0q1, p1q1, a);

    float b = cubeSmooth(fract(pos.y*texSize.y));

    return mix(pInterp_q0, pInterp_q1, b);
}

float mieHG(float cosTheta, float g) {
    float mie   = 1.0 + sqr(g) - 2.0*g*cosTheta;
        mie     = (1.0 - sqr(g)) / ((4.0*pi) * mie*(mie*0.5+0.5));
    return mie;
}

float cloudPhase(float cosTheta, float g, vec3 gMult) {
    float x = mieHG(cosTheta, gMult.x * g);
    float y = mieHG(cosTheta, -gMult.y * g);
    float z = mieHG(cosTheta, gMult.z * g) * sqrt2;

    return mix(mix(x, y, 0.2), z, 0.15);    //i assume this is more energy conserving than summing them
}

const float cloudScale  = 0.17 / 1024.0;
const float cloudVolMaxY   = float(cloudVolumeAlt + cloudVolumeDepth);
const float cloudVolMidY   = float(cloudVolumeAlt) + cloudVolumeDepth * 0.5;

float cloudVolumeShape(vec3 pos) {
    float fadeLow   = sstep(pos.y, float(cloudVolumeAlt), float(cloudVolumeAlt) + float(cloudVolumeDepth) * 0.02);
    float fadeHigh  = 1.0 - sstep(pos.y, cloudVolMaxY - float(cloudVolumeDepth) * 0.02, cloudVolMaxY);
    float erodeLow  = cube(1.0 - linStep(pos.y, float(cloudVolumeAlt), float(cloudVolumeAlt) + float(cloudVolumeDepth) * 0.2));
    float erodeHigh = cube(linStep(pos.y, float(cloudVolumeAlt) + float(cloudVolumeDepth) * 0.8, cloudVolMaxY));

    float altitude  = pos.y;

        #ifdef volumeWorldTimeAnim
            pos.x  += worldAnimTime * 600.0;
        #else
            pos.x  += frameTimeCounter;
        #endif

        pos    *= cloudScale;

    float shape = cloudTextureCubic(colortex8, pos.xz);
        shape  -= erodeLow + erodeHigh;
        shape  *= fadeLow * fadeHigh;
        shape  -= 0.007;

    #ifdef cloudVolumeStoryMode
    float storyFade = linStep(altitude, cloudVolumeAlt + 2.0, cloudVolMaxY);
        shape  *= cube(1.0 - storyFade) * 0.99 + 0.01;
    #endif

    return max0(shape);
}

float cloudVolumeLightOD(vec3 pos, const uint steps, vec3 dir) {
    float stepSize      = float(cloudVolumeDepth) / float(steps);

    vec3 rStep  = dir * stepSize;

        pos    += rStep / pi;

    float od    = 0.0;

    for (uint i = 0; i < steps; ++i, pos += rStep) {

        if(pos.y > cloudVolMaxY || pos.y < cloudVolumeAlt) continue;

        float density   = cloudVolumeShape(pos);

            od += density * stepSize;
    }

    return od;
}

float approxSkylightDensity(vec3 pos) {
    float distToTop     = max0(cloudVolMaxY - pos.y);

    return distToTop;
}

vec4 volumetricClouds(vec3 worldDir, vec3 worldPos, float vDotL, float dither, bool terrain, vec3 skyColor) {
    vec3 totalScattering    = vec3(0.0);
    float totalTransmittance = 1.0;

    vec3 sunlight       = (worldTime>23000 || worldTime<12900) ? lightColor[0] : lightColor[1];
        sunlight       *= cloudLightFlip * sqrt2;
    vec3 skylight       = lightColor[2] * (1.0 - sqrt(daytime.w)*0.96);

    float pFade         = saturate(mieHG(vDotL, 0.65));

    float within    = sstep(eyeAltitude, float(cloudVolumeAlt) - 15.0, float(cloudVolumeAlt)) * (1.0 - sstep(eyeAltitude, cloudVolMaxY, cloudVolMaxY + 15.0));
    bool isBelowVol = eyeAltitude < cloudVolMidY;
    bool visibleVol = worldDir.y > 0.0 && isBelowVol || worldDir.y < 0.0 && !isBelowVol;

    const float sigmaA  = 1.0;
    const float sigmaT  = 0.66;

    if (visibleVol || within > 0.0) {
        vec3 bottom     = worldDir * ((cloudVolumeAlt - eyeAltitude) * rcp(worldDir.y));
        vec3 top        = worldDir * ((cloudVolMaxY - eyeAltitude) * rcp(worldDir.y));

        if (worldDir.y < 0.0 && isBelowVol || worldDir.y > 0.0 && !isBelowVol) {
            bottom      = vec3(0.0);
            top         = vec3(0.0);
        }

        vec3 start      = isBelowVol ? bottom : top;
        vec3 end        = isBelowVol ? top : bottom;
            start       = mix(start, gbufferModelViewInverse[3].xyz, within);
            end         = mix(end, worldDir * cloudVolumeClip, within);

        if (terrain) end = worldPos;

        float stepCoeff     = 1.0 + sqr(max(within, float(terrain))) * 4.0;
        uint steps          = uint(cloudVolumeSamples * stepCoeff);

        vec3 rStep          = (end - start) * rcp(float(steps));
        vec3 rPos           = rStep * dither + start + cameraPosition;
        float rLength       = length(rStep);

        vec3 scattering     = vec3(0.0);
        float transmittance = 1.0;

        for (uint i = 0; i < steps; ++i, rPos += rStep) {
            if (transmittance < 0.01) break;
            if (rPos.y < cloudVolumeAlt || rPos.y > cloudVolMaxY) continue;

            float dist  = distance(rPos, cameraPosition);
            if (dist > cloudVolumeClip) continue;

            float density = cloudVolumeShape(rPos);
            if (density <= 0.0) continue;

            float extinction    = density * sigmaT;
            float stepT         = exp(-extinction * rLength);
            float integral      = (1.0 - stepT) * rcp(sigmaT);

            vec3 stepScatter    = vec3(0.0);

            float lightOD       = cloudVolumeLightOD(rPos, 5, cloudLightDir) * sigmaA;
            float skyOD         = approxSkylightDensity(rPos) * sigmaA;

            float powder        = 8.0 * (1.0 - 0.97 * exp(-extinction * 18.0));

            float anisoPowder   = mix(powder, 1.0, pFade);

            vec3 phaseG         = pow(vec3(0.45, 0.25, 0.95), vec3(1.0 + lightOD));

            float phase = cloudPhase(vDotL, 1.0, phaseG);

            stepScatter.x  += max(expf(-lightOD * sigmaT), expf(-lightOD * sigmaT * 0.2) * 0.75) * phase * anisoPowder * sigmaT;
            stepScatter.y  += max(expf(-skyOD * sigmaT), expf(-skyOD * sigmaT * 0.2) * 0.75) * powder * sigmaT;

            stepScatter     = (sunlight * stepScatter.x) + (skylight * stepScatter.y);

            float atmosFade = expf(-dist * 4e-3);

            stepScatter     = mix(skyColor * sigmaT, stepScatter, atmosFade);

            scattering     += stepScatter * (integral * transmittance);

            transmittance  *= stepT;
        }

        transmittance       = linStep(transmittance, 0.01, 1.0);

        totalScattering    += scattering;
        totalTransmittance *= transmittance;
    }

    vec4 result     = vec4(totalScattering, totalTransmittance);
        result      = mix(result, vec4(0,0,0,1), exp(-max0(worldDir.y * pi4)) * (1.0 - within) * float(eyeAltitude < cloudVolumeAlt));

    return result;
}

void main() {
    sceneColor  = texture(colortex0, uv).rgb;

    vec4 translucencyColor  = texture(colortex5, uv);

    vec2 sceneDepth     = vec2(texture(depthtex0, uv).x, texture(depthtex1, uv).x);

    vec3 position0      = vec3(uv, sceneDepth.x);
        position0       = screenToViewSpace(position0);

    vec3 position1      = vec3(uv, sceneDepth.y);
        position1       = screenToViewSpace(position1);

    mat2x3 scenePos     = mat2x3(viewToSceneSpace(position0), viewToSceneSpace(position1));

    vec3 worldDir       = mat3(gbufferModelViewInverse) * normalize(position0);

    vec3 skyColor       = texture(colortex3, projectSkyTiles(worldDir) * vec2(1.0, 0.5), 0).rgb;

    #ifdef fogEnabled
    sceneColor  = getFog(sceneColor, distance(position0, position1), skyColor, worldDir);
    #endif

    sceneColor  = sceneColor * sqr(texture(colortex7, uv).rgb) * (1.0 - translucencyColor.a) + translucencyColor.rgb;

    #ifdef fogEnabled
    sceneColor  = getFog(sceneColor, length(position0), skyColor, worldDir);
    #endif

    #ifdef cloudVolumeEnabled
        vec4 cloudData  = volumetricClouds(worldDir, scenePos[0], dot(normalize(position1), cloudLightDirView), ditherBluenoise(), landMask(sceneDepth.y), skyColor);

    sceneColor  = sceneColor * cloudData.a + cloudData.rgb;
    #endif
}